# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import logging

import sys
import os
import pytest
from pyspark.sql import SparkSession


sys.path.append(os.path.join(os.path.dirname(__file__), 'helpers'))

_spark_session = None
_spark_context = None


def pytest_configure(config):
    global _spark_session, _spark_context

    _spark_session = (
        SparkSession.builder.master("local").appName("mozanalysis_test").getOrCreate()
    )
    _spark_context = _spark_session.sparkContext

    logger = logging.getLogger("py4j")
    logger.setLevel(logging.ERROR)


def pytest_unconfigure(config):
    _spark_session.stop()


@pytest.fixture()
def spark():
    return _spark_session


@pytest.fixture()
def spark_context(spark):
    return _spark_context


def pytest_generate_tests(metafunc):
    if 'spark_context_or_none' in metafunc.fixturenames:
        metafunc.parametrize(
            "spark_context_or_none", ["spark", "no spark"], indirect=True
        )


@pytest.fixture()
def spark_context_or_none(request):
    if request.param == 'spark':
        return _spark_context
    elif request.param == 'no spark':
        return None
    else:
        raise ValueError("invalid internal test config")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import re
from google.cloud import bigquery
from google.api_core.exceptions import Conflict


def sanitize_table_name_for_bq(table_name):
    of_good_character_but_possibly_verbose = re.sub(r'[^a-zA-Z_0-9]', '_', table_name)

    if len(of_good_character_but_possibly_verbose) <= 1024:
        return of_good_character_but_possibly_verbose

    return of_good_character_but_possibly_verbose[:500] + '___' \
        + of_good_character_but_possibly_verbose[-500:]


class BigQueryContext:
    """Holds a BigQuery client, and some configuration.

    Args:
        dataset_id (str): Your `BigQuery dataset id`_.
        project_id (str, optional): Your BigQuery project_id.
            Defaults to the DS team's project.

    .. _BigQuery dataset id: https://cloud.google.com/bigquery/docs/datasets
    """
    def __init__(self, dataset_id, project_id='moz-fx-data-bq-data-science'):
        self.dataset_id = dataset_id
        self.project_id = project_id
        self.client = bigquery.Client(project=project_id)

    def run_query(self, sql, results_table=None):
        """Run a query and return the result.

        If ``results_table`` is provided, then save the results
        into there (or just query from there if it already exists).

        Returns a ``google.cloud.bigquery.table.RowIterator``
        """
        if not results_table:
            return self.client.query(sql).result()

        try:
            full_res = self.client.query(
                sql,
                job_config=bigquery.QueryJobConfig(
                    destination=self.client.dataset(
                        self.dataset_id
                    ).table(results_table)
                )
            ).result()
            print('Saved into', results_table)
            return full_res

        except Conflict:
            print("Full results table already exists. Reusing", results_table)
            return self.client.query(
                "SELECT * FROM {}".format(
                    self.fully_qualify_table_name(results_table)
                )
            ).result()

    def fully_qualify_table_name(self, table_name):
        """Given a table name, return it fully qualified."""
        return "`{project_id}.{dataset_id}.{full_table_name}`".format(
            project_id=self.project_id,
            dataset_id=self.dataset_id,
            full_table_name=table_name,
        )



import pytest
from cheap_lint import sql_lint

import mozanalysis.segments.desktop as msd
import mozanalysis.metrics.desktop as mmd


def test_imported_ok():
    assert msd.regular_users_v2


def test_sql_not_detectably_malformed():
    for m in msd.__dict__.values():
        if isinstance(m, msd.Segment):
            sql_lint(m.select_expr)

    for ds in msd.__dict__.values():
        if isinstance(ds, msd.SegmentDataSource):
            sql_lint(ds.from_expr)


def test_consistency_of_segment_and_variable_names():
    for name, segment in msd.__dict__.items():
        if isinstance(segment, msd.Segment):
            assert name == segment.name, segment


def test_segment_data_source_window_end_validates():
    msd.SegmentDataSource(
        name='bla',
        from_expr="bla",
        window_start=0,
        window_end=0,
    )

    with pytest.raises(ValueError):
        msd.SegmentDataSource(
            name='bla',
            from_expr="bla",
            window_start=0,
            window_end=1,
        )


def test_segment_data_source_window_start_validates():
    msd.SegmentDataSource(
        name='bla',
        from_expr="bla",
        window_start=-1,
        window_end=-1,
    )

    with pytest.raises(ValueError):
        msd.SegmentDataSource(
            name='bla',
            from_expr="bla",
            window_start=0,
            window_end=-1,
        )


def test_segment_validates_not_metric_data_source():
    with pytest.raises(TypeError):
        msd.Segment(
            name='bla',
            data_source=mmd.clients_daily,
            select_expr='bla',
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozanalysis.metrics import Metric, DataSource, agg_sum, agg_any


clients_daily = DataSource(
    name='clients_daily',
    from_expr="`moz-fx-data-shared-prod.telemetry.clients_daily`",
)

search_clients_daily = DataSource(
    name='search_clients_daily',
    from_expr='`moz-fx-data-shared-prod.search.search_clients_daily`',
    experiments_column_type=None,
)

main_summary = DataSource(
    name='main_summary',
    from_expr="`moz-fx-data-shared-prod.telemetry.main_summary`"
)

events = DataSource(
    name='events',
    from_expr="`moz-fx-data-shared-prod.telemetry.events`",
    experiments_column_type='native',
)

main = DataSource(
    name='main',
    from_expr="""(
                SELECT
                    *,
                    DATE(submission_timestamp) AS submission_date,
                    environment.experiments
                FROM `moz-fx-data-shared-prod`.telemetry.main
            )""",
    experiments_column_type="native",
)

crash = DataSource(
    name='crash',
    from_expr="""(
                SELECT
                    *,
                    DATE(submission_timestamp) AS submission_date,
                    environment.experiments
                FROM `moz-fx-data-shared-prod`.telemetry.crash
            )""",
    experiments_column_type="native",
)

active_hours = Metric(
    name='active_hours',
    data_source=clients_daily,
    select_expr=agg_sum('active_hours_sum')
)

uri_count = Metric(
    name='uri_count',
    data_source=clients_daily,
    select_expr=agg_sum('scalar_parent_browser_engagement_total_uri_count_sum')
)

search_count = Metric(
    name='search_count',
    data_source=search_clients_daily,
    select_expr=agg_sum('sap')
)

tagged_search_count = Metric(
    name='tagged_search_count',
    data_source=search_clients_daily,
    select_expr=agg_sum('tagged_sap')
)

tagged_follow_on_search_count = Metric(
    name='tagged_follow_on_search_count',
    data_source=search_clients_daily,
    select_expr=agg_sum('tagged_follow_on')
)

ad_clicks = Metric(
    name='ad_clicks',
    data_source=search_clients_daily,
    select_expr=agg_sum('ad_click')
)

searches_with_ads = Metric(
    name='searches_with_ads',
    data_source=search_clients_daily,
    select_expr=agg_sum('search_with_ads')
)

organic_search_count = Metric(
    name='organic_search_count',
    data_source=search_clients_daily,
    select_expr=agg_sum('organic')
)

unenroll = Metric(
    name='unenroll',
    data_source=events,
    select_expr=agg_any("""
                event_category = 'normandy'
                AND event_method = 'unenroll'
                AND event_string_value = '{experiment_slug}'
            """)
)

view_about_logins = Metric(
    name='view_about_logins',
    data_source=events,
    select_expr=agg_any("""
                event_method = 'open_management'
                AND event_category = 'pwmgr'
            """)
)

view_about_protections = Metric(
    name='view_about_protections',
    data_source=events,
    select_expr=agg_any("""
                event_method = 'show'
                AND event_object = 'protection_report'
            """)
)

connect_fxa = Metric(
    name='connect_fxa',
    data_source=events,
    select_expr=agg_any("""
                event_method = 'connect'
                AND event_object = 'account'
            """)
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pandas as pd
import pytest

import mozanalysis.bayesian_stats.binary as mabsb
import mozanalysis.bayesian_stats as mabs


def test_aggregate_col():
    df = pd.DataFrame([
        ['ctl', 1, 0],
        ['ctl', 1, 0],
        ['ctl', 0, 1],
        ['tst', 1, 0],
        ['tst', 0, 1],
    ], columns=['branch', 'val1', 'val2'])

    res1 = mabsb.aggregate_col(df, 'val1')
    assert res1.loc['ctl', 'num_enrollments'] == 3
    assert res1.loc['tst', 'num_enrollments'] == 2
    assert res1.loc['ctl', 'num_conversions'] == 2
    assert res1.loc['tst', 'num_conversions'] == 1

    res2 = mabsb.aggregate_col(df, 'val2')
    assert res2.loc['ctl', 'num_enrollments'] == 3
    assert res2.loc['tst', 'num_enrollments'] == 2
    assert res2.loc['ctl', 'num_conversions'] == 1
    assert res2.loc['tst', 'num_conversions'] == 1


def test_summarize_one_branch_vs_samples():
    s = pd.Series([30, 80], index=['num_conversions', 'num_enrollments'])

    ppf = mabsb.summarize_one_branch_from_agg(s, quantiles=[0.5, 0.41])

    samples = mabsb.get_samples(
        s.to_frame().T, 'num_enrollments', 'num_conversions', 100000
    )

    res = mabs.summarize_one_branch_samples(samples.iloc[:, 0], quantiles=[0.5, 0.41])

    assert ppf['0.5'] == pytest.approx(res['0.5'], abs=0.001)
    assert ppf['0.41'] == pytest.approx(res['0.41'], abs=0.001)
    assert ppf['mean'] == pytest.approx(res['mean'], abs=0.001)


def test_compare_branches():
    # The stats has been checked above; this just exercises the remaining
    # business logic

    df = pd.DataFrame([
        ['ctl', 1, 0],
        ['ctl', 1, 0],
        ['ctl', 0, 1],
        ['tst', 1, 0],
        ['tst', 0, 1],
        ['3rd', 1, 0],
        ['3rd', 0, 1],
    ], columns=['branch', 'val1', 'val2'])

    res = mabsb.compare_branches(df, 'val2', ref_branch_label='ctl')

    assert set(res['individual'].keys()) == set(df.branch.unique())
    assert set(res['comparative'].keys()) == set(df.branch.unique()) - set(['ctl'])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd

import mozanalysis.bayesian_stats as mabs
from mozanalysis.utils import filter_outliers


def bb_mean(values, prob_weights):
    """Calculate the mean of a bootstrap replicate.

    Args:
        values (pd.Series, ndarray): One dimensional array
            of observed values
        prob_weights (pd.Series, ndarray): Equally shaped
            array of the probability weight associated with
            each value.

    Returns:
        The mean as a np.float.
    """
    return np.dot(values, prob_weights)


def make_bb_quantile_closure(quantiles):
    """Return a function to calculate quantiles for a bootstrap replicate.

    Args:
        quantiles (float, list of floats): Quantiles to compute

    Returns a function that calculates quantiles for a bootstrap replicate:

        Args:

            values (pd.Series, ndarray):
                One dimensional array of observed values
            prob_weights (pd.Series, ndarray):
                Equally shaped array of the probability weight associated with
                each value.

        Returns:

            * A quantile as a np.float, or
            * several quantiles as a dict keyed by the quantiles

    """

    # If https://github.com/numpy/numpy/pull/9211/ is ever merged then
    # we can just use that instead.

    def get_value_at_quantile(values, cdf, quantile):
        """Return the value at a quantile.

        Does no interpolation because our Bayesian bootstrap
        implementation calls `np.unique` to tally the values:
        if it did not take this shortcut then regardless of whether
        we interpolate when returning quantiles, the vast majority
        of quantiles would coincide with a value. But since we take
        this shortcut, interpolation mostly returns values not in the
        dataset. Ergh.
        """
        # Add a tolerance of 1e-6 to account for numerical error when
        # computing the cdf
        arg = np.nonzero(quantile < cdf + 1e-6)[0][0]
        return values[arg]

    def bb_quantile(values, prob_weights):
        """Calculate quantiles for a bootstrap replicate.

        Args:
            values (pd.Series, ndarray): One dimensional array
                of observed values
            prob_weights (pd.Series, ndarray): Equally shaped
                array of the probability weight associated with
                each value.

        Returns:

            * A quantile as a np.float, or
            * several quantiles as a dict keyed by the quantiles
        """
        # assume values is previously sorted, as per np.unique()
        cdf = np.cumsum(prob_weights)

        if np.isscalar(quantiles):
            return get_value_at_quantile(values, cdf, quantiles)

        else:
            return {
                q: get_value_at_quantile(values, cdf, q)
                for q in quantiles
            }

    return bb_quantile


def compare_branches(
    df, col_label, ref_branch_label='control', stat_fn=bb_mean,
    num_samples=10000, threshold_quantile=None,
    individual_summary_quantiles=mabs.DEFAULT_QUANTILES,
    comparative_summary_quantiles=mabs.DEFAULT_QUANTILES,
    sc=None
):
    """Jointly sample bootstrapped statistics then compare them.

    Args:
        df: a pandas DataFrame of queried experiment data in the
            standard format (see `mozanalysis.experiment`).
        col_label (str): Label for the df column contaning the metric
            to be analyzed.
        ref_branch_label (str, optional): String in ``df['branch']``
            that identifies the branch with respect to which we want to
            calculate uplifts - usually the control branch.
        stat_fn (callable, optional): A function that either:

            * Aggregates each resampled population to a scalar (e.g.
              the default, ``bb_mean``), or
            * Aggregates each resampled population to a dict of
              scalars (e.g. the func returned by
              ``make_bb_quantile_closure`` when given multiple
              quantiles.

            In both cases, this function must accept two parameters:

            * a one-dimensional ndarray or pandas Series of values,
            * an identically shaped object of weights for these values

        num_samples (int, optional): The number of bootstrap iterations
            to perform.
        threshold_quantile (float, optional): An optional threshold
            quantile, above which to discard outliers. E.g. `0.9999`.
        individual_summary_quantiles (list, optional): Quantiles to
            determine the credible intervals on individual branch
            statistics. Change these when making Bonferroni corrections.
        comparative_summary_quantiles (list, optional): Quantiles to
            determine the credible intervals on comparative branch
            statistics (i.e. the change relative to the reference
            branch, probably the control). Change these when making
            Bonferroni corrections.
        sc (optional): The Spark context, if available

    Returns:
        If ``stat_fn`` returns a scalar (this is the default), then
        this function returns a dictionary has the following keys and
        values:

            * 'individual': dictionary mapping each branch name to a pandas
              Series that holds the expected value for the bootstrapped
              ``stat_fn``, and credible intervals.
            * 'comparative': dictionary mapping each branch name to a pandas
              Series of summary statistics for the possible uplifts of
              the bootstrapped ``stat_fn`` relative to the reference branch.

        Otherwise, when ``stat_fn`` returns a dict, then this function
        returns a similar dictionary, except the Series are replaced with
        DataFrames. Each row in each DataFrame corresponds to one output
        of ``stat_fn``, and is the Series that would be returned if
        ``stat_fn`` computed only this statistic.
    """
    branch_list = df.branch.unique()

    if ref_branch_label not in branch_list:
        raise ValueError("Branch label '{b}' not in branch list '{bl}".format(
            b=ref_branch_label, bl=branch_list
        ))

    samples = {
        # TODO: do we need to control seed_start? If so then we must be careful here
        b: get_bootstrap_samples(
            df[col_label][df.branch == b],
            stat_fn,
            num_samples,
            threshold_quantile=threshold_quantile,
            sc=sc,
        ) for b in branch_list
    }

    return mabs.compare_samples(
        samples, ref_branch_label, individual_summary_quantiles,
        comparative_summary_quantiles
    )


def bootstrap_one_branch(
    data, stat_fn=bb_mean, num_samples=10000, seed_start=None,
    threshold_quantile=None, summary_quantiles=mabs.DEFAULT_QUANTILES,
    sc=None
):
    """Bootstrap ``stat_fn`` for one branch on its own.

    Computes ``stat_fn`` for ``num_samples`` resamples of ``data``,
    then returns summary statistics for the results.

    Args:
        data: The data as a list, 1D numpy array, or pandas Series
        stat_fn (callable, optional): A function that either:

            * Aggregates each resampled population to a scalar (e.g.
              the default, ``bb_mean``), or
            * Aggregates each resampled population to a dict of
              scalars (e.g. the func returned by
              ``make_bb_quantile_closure`` when given multiple
              quantiles.

            In both cases, this function must accept two parameters:

            * a one-dimensional ndarray or pandas Series of values,
            * an identically shaped object of weights for these values

        num_samples: The number of bootstrap iterations to perform
        seed_start: An int with which to seed numpy's RNG. It must
            be unique within this set of calculations.
        threshold_quantile (float, optional): An optional threshold
            quantile, above which to discard outliers. E.g. ``0.9999``.
        summary_quantiles (list, optional): Quantiles to determine the
            confidence bands on the branch statistics. Change these
            when making Bonferroni corrections.
        sc (optional): The Spark context, if available
    """
    samples = get_bootstrap_samples(
        data, stat_fn, num_samples, seed_start, threshold_quantile, sc
    )

    return mabs.summarize_one_branch_samples(samples, summary_quantiles)


def get_bootstrap_samples(
    data, stat_fn=bb_mean, num_samples=10000, seed_start=None,
    threshold_quantile=None, sc=None
):
    """Return ``stat_fn`` evaluated on resampled data.

    Args:
        data: The data as a list, 1D numpy array, or pandas series
        stat_fn (callable, optional): A function that either:

            * Aggregates each resampled population to a scalar (e.g.
              the default, ``bb_mean``), or
            * Aggregates each resampled population to a dict of
              scalars (e.g. the func returned by
              ``make_bb_quantile_closure`` when given multiple
              quantiles.

            In both cases, this function must accept two parameters:

            * a one-dimensional ndarray or pandas Series of values,
            * an identically shaped object of weights for these values

        num_samples: The number of samples to return
        seed_start: A seed for the random number generator; this
            function will use seeds in the range::

                [seed_start, seed_start + num_samples)

            and these particular seeds must not be used elsewhere
            in this calculation. By default, use a random seed.

        threshold_quantile (float, optional): An optional threshold
            quantile, above which to discard outliers. E.g. ``0.9999``.
        sc (optional): The Spark context, if available

    Returns:
        A Series or DataFrame with one row per sample and one column
        per output of ``stat_fn``.

    References:
        Rubin, Donald B. The Bayesian Bootstrap. Ann. Statist. 9 (1981),
            no. 1, 130--134. https://dx.doi.org/10.1214/aos/1176345338
    """
    if not type(data) == np.ndarray:
        data = np.array(data)

    if np.isnan(data).any():
        raise ValueError("'data' contains null values")

    if threshold_quantile:
        data = filter_outliers(data, threshold_quantile)

    # For computational efficiency, tally the data. If you are careful
    # with the resulting draws from the dirichlet then this should be
    # equivalent to not doing this step (and passing np.ones() as the
    # counts)
    data_values, data_counts = np.unique(data, return_counts=True)

    if seed_start is None:
        seed_start = np.random.randint(np.iinfo(np.uint32).max)

    # Deterministic "randomness" requires careful state handling :(
    # Need to ensure every call has a unique, deterministic seed.
    seed_range = range(seed_start, seed_start + num_samples)

    if sc is None:
        summary_stat_samples = [
            _resample_and_agg_once(data_values, data_counts, stat_fn, unique_seed)
            for unique_seed in seed_range
        ]

    else:
        try:
            broadcast_data_values = sc.broadcast(data_values)
            broadcast_data_counts = sc.broadcast(data_counts)

            summary_stat_samples = sc.parallelize(seed_range).map(
                lambda seed: _resample_and_agg_once_bcast(
                    broadcast_data_values=broadcast_data_values,
                    broadcast_data_counts=broadcast_data_counts,
                    stat_fn=stat_fn,
                    unique_seed=seed % np.iinfo(np.uint32).max,
                )
            ).collect()

        finally:
            broadcast_data_values.unpersist()
            broadcast_data_counts.unpersist()

    summary_df = pd.DataFrame(summary_stat_samples)
    if len(summary_df.columns) == 1:
        # Return a Series if stat_fn returns a scalar
        return summary_df.iloc[:, 0]

    # Else return a DataFrame if stat_fn returns a dict
    return summary_df


def _resample_and_agg_once_bcast(
    broadcast_data_values, broadcast_data_counts, stat_fn, unique_seed
):
    return _resample_and_agg_once(
        broadcast_data_values.value, broadcast_data_counts.value,
        stat_fn, unique_seed
    )


def _resample_and_agg_once(
    data_values, data_counts, stat_fn, unique_seed=None
):
    random_state = np.random.RandomState(unique_seed)

    prob_weights = random_state.dirichlet(data_counts)

    return stat_fn(data_values, prob_weights)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozanalysis.segments import Segment, SegmentDataSource


clients_last_seen_lag_1_day = SegmentDataSource(
    name='clients_last_seen',
    from_expr="`moz-fx-data-shared-prod.telemetry.clients_last_seen`",
    window_start=0,
    window_end=0,
)


# Before https://github.com/mozilla/bigquery-etl/pull/825/ is merged,
# we gotta do our own transforms
tmp_segment_regular_users_v2 = """  CASE
  WHEN
    BIT_COUNT(days_visited_5_uri_bits & 0x0FFFFFFE) = 0
  THEN
    'new_irregular_users_v2'
  WHEN
    BIT_COUNT(days_visited_5_uri_bits & 0x0FFFFFFE)
    BETWEEN 1
    AND 7
  THEN
    'semi_regular_users_v2'
  WHEN
    BIT_COUNT(days_visited_5_uri_bits & 0x0FFFFFFE)
    BETWEEN 8
    AND 27
  THEN
    'regular_users_v2'
  END
"""


regular_users_v2 = Segment(
    name='regular_users_v2',
    data_source=clients_last_seen_lag_1_day,
    # select_expr="""MAX(
    #     COALESCE(segment_regular_users_v2, 'new_irregular_users_v2')
    #     = 'regular_users_v2'
    # )""",
    select_expr=f"""MAX(
        COALESCE({tmp_segment_regular_users_v2}, 'new_irregular_users_v2')
        = 'regular_users_v2'
    )""",
)

semi_regular_users_v2 = Segment(
    name='semi_regular_users_v2',
    data_source=clients_last_seen_lag_1_day,
    # select_expr="""MAX(
    #     COALESCE(segment_regular_users_v2, 'new_irregular_users_v2')
    #     = 'semi_regular_users_v2'
    # )""",
    select_expr=f"""MAX(
        COALESCE({tmp_segment_regular_users_v2}, 'new_irregular_users_v2')
        = 'semi_regular_users_v2'
    )""",
)

new_irregular_users_v2 = Segment(
    name='new_irregular_users_v2',
    data_source=clients_last_seen_lag_1_day,
    # select_expr="""MAX(
    #     COALESCE(segment_regular_users_v2, 'new_irregular_users_v2')
    #     = 'new_irregular_users_v2'
    # )""",
    select_expr=f"""MAX(
        COALESCE({tmp_segment_regular_users_v2}, 'new_irregular_users_v2')
        = 'new_irregular_users_v2'
    )""",
)



from cheap_lint import sql_lint

import mozanalysis.metrics.desktop as mmd


def test_imported_ok():
    assert mmd.active_hours


def test_sql_not_detectably_malformed():
    for m in mmd.__dict__.values():
        if isinstance(m, mmd.Metric):
            sql_lint(m.select_expr.format(experiment_slug='slug'))

    for ds in mmd.__dict__.values():
        if isinstance(ds, mmd.DataSource):
            sql_lint(ds.from_expr)


def test_consistency_of_metric_and_variable_names():
    for name, metric in mmd.__dict__.items():
        if isinstance(metric, mmd.Metric):
            assert name == metric.name, metric



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pandas as pd
import numpy as np
import scipy.stats as st

import mozanalysis.bayesian_stats as mabs


def compare_branches(
    df, col_label, ref_branch_label='control', num_samples=10000,
    individual_summary_quantiles=mabs.DEFAULT_QUANTILES,
    comparative_summary_quantiles=mabs.DEFAULT_QUANTILES
):
    """Jointly sample conversion rates for branches then compare them.

    See `compare_branches_from_agg` for more details.

    Args:
        df (pd.DataFrame): Queried experiment data in the standard
            format.
        col_label (str): Label for the df column contaning the metric
            to be analyzed.
        ref_branch_label (str, optional): String in ``df['branch']``
            that identifies the the branch with respect to which we
            want to calculate uplifts - usually the control branch.
        num_samples (int, optional): The number of samples to compute.
        individual_summary_quantiles (list, optional): Quantiles to
            determine the confidence bands on individual branch
            statistics. Change these when making Bonferroni corrections.
        comparative_summary_quantiles (list, optional): Quantiles to
            determine the confidence bands on comparative branch
            statistics (i.e. the change relative to the reference
            branch, probably the control). Change these when making
            Bonferroni corrections.

    Returns a dictionary:

        * 'individual': dictionary mapping branch names to a pandas
          Series of summary stats for the posterior distribution over
          the branch's conversion rate.
        * 'comparative': dictionary mapping branch names to a pandas
          Series of summary statistics for the possible uplifts of the
          conversion rate relative to the reference branch - see docs
          for
          :meth:`mozanalysis.stats.summarize_samples.summarize_joint_samples`.
    """
    agg_col = aggregate_col(df, col_label)

    return compare_branches_from_agg(
        agg_col, ref_branch_label=ref_branch_label, num_samples=num_samples,
        individual_summary_quantiles=mabs.DEFAULT_QUANTILES,
        comparative_summary_quantiles=mabs.DEFAULT_QUANTILES
    )


def aggregate_col(df, col_label):
    """Return the number of enrollments and conversions per branch.

    Args:
        df (pd.DataFrame): Queried experiment data in the standard
            format.
        col_label (str): Label for the df column contaning the metric
            to be analyzed.

    Returns:
        A DataFrame. The index is the list of branches. It has the
        following columns:

        * num_enrollments: The number of experiment subjects enrolled in
          this branch who were eligible for the metric.
        * num_conversions: The number of these enrolled experiment subjects
          who met the metric's conversion criteria.
    """
    # I would have used `isin` but it seems to be ~100x slower?
    if not ((df[col_label] == 0) | (df[col_label] == 1)).all():
        raise ValueError("All values in column '{}' must be 0 or 1.".format(col_label))

    return df.groupby('branch')[col_label].agg(
        ['count', 'sum']
    ).rename(
        columns={'count': 'num_enrollments', 'sum': 'num_conversions'}
    )


def summarize_one_branch_from_agg(
    s, num_enrollments_label='num_enrollments', num_conversions_label='num_conversions',
    quantiles=mabs.DEFAULT_QUANTILES
):
    """Return stats about a branch's conversion rate.

    Calculate and return a Series of summary stats for the posterior
    distribution over the branch's conversion rate.

    Args:
        s (pd.Series): Holds the number of enrollments and number of
            conversions for this branch and metric.
        num_enrollments_label (str, optional): The label in this Series
            for the number of enrollments
        num_conversions_label (str, optional): The label in this Series
            for the number of conversions
        quantiles (list, optional): The quantiles to return as summary
            statistics.

    Returns:
        A pandas Series; the index contains the stringified
        ``quantiles`` plus ``'mean'``.
    """
    beta = st.beta(
        s.loc[num_conversions_label] + 1,
        s.loc[num_enrollments_label] - s.loc[num_conversions_label] + 1
    )

    q_index = [str(v) for v in quantiles]

    res = pd.Series(index=q_index + ['mean'])

    res[q_index] = beta.ppf(quantiles)
    res['mean'] = beta.mean()

    return res


def compare_branches_from_agg(
    df,
    ref_branch_label='control',
    num_enrollments_label='num_enrollments',
    num_conversions_label='num_conversions',
    num_samples=10000,
    individual_summary_quantiles=mabs.DEFAULT_QUANTILES,
    comparative_summary_quantiles=mabs.DEFAULT_QUANTILES
):
    """Jointly sample conversion rates for two branches then compare them.

    Calculates various quantiles on the uplift of the non-control
    branch's sampled conversion rates with respect to the control
    branch's sampled conversion rates.

    The data in `df` is modelled as being generated binomially, with a
    Beta(1, 1) (uniform) prior over the conversion rate parameter.

    Args:
        df: A pandas dataframe of integers.

            * ``df.index`` lists the experiment branches
            * ``df.columns`` is
              ``[num_enrollments_label, num_conversions_label]``

        ref_branch_label (str, optional): Label for the df row
            containing data for the control branch
        num_enrollments_label: Label for the df column containing the
            number of enrollments in each branch.
        num_conversions_label: Label for the df column containing the
            number of conversions in each branch.
        num_samples: The number of samples to compute

    Returns a dictionary:

        * 'individual': dictionary mapping branch names to a pandas
          Series of summary stats for the posterior distribution over
          the branch's conversion rate.
        * 'comparative': dictionary mapping branch names to a pandas
          Series of summary statistics for the possible uplifts of the
          conversion rate relative to the reference branch - see docs
          for
          :meth:`mozanalysis.stats.summarize_samples.summarize_joint_samples`.
    """
    assert ref_branch_label in df.index, "What's the reference branch?"

    samples = get_samples(
        df, num_enrollments_label, num_conversions_label, num_samples
    )

    return {
        'individual': {
                b: summarize_one_branch_from_agg(
                    df.loc[b], num_enrollments_label, num_conversions_label,
                    quantiles=individual_summary_quantiles
                ) for b in df.index
            },
        'comparative': {
                b: mabs.summarize_joint_samples(
                    samples[b], samples[ref_branch_label],
                    quantiles=comparative_summary_quantiles
                ) for b in df.index.drop(ref_branch_label)
            },
    }


def get_samples(
    df, num_enrollments_label, num_conversions_label, num_samples
):
    """Return samples from Beta distributions.

    Assumes a Beta(1, 1) prior.

    Args:
        df: A pandas dataframe of integers:

            * ``df.index`` lists the experiment branches
            * ``df.columns`` is
              ``(num_enrollments_label, num_conversions_label)``

        num_enrollments_label: Label for the df column containing the
            number of enrollments in each branch.
        num_conversions_label: Label for the df column containing the
            number of conversions in each branch.
        num_samples: The number of samples to compute

    Returns a pandas.DataFrame of sampled conversion rates

        * columns: list of branches
        * index: enumeration of samples
    """
    samples = pd.DataFrame(index=np.arange(num_samples), columns=df.index)
    for branch_label, r in df.iterrows():
        # Oh, for a better prior...
        samples[branch_label] = np.random.beta(
            r.loc[num_conversions_label] + 1,
            r.loc[num_enrollments_label] - r.loc[num_conversions_label] + 1,
            size=num_samples
        )

    return samples



#!/usr/bin/env python
# encoding: utf-8

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from setuptools import setup, find_packages

tests_require = [
    'mock', 'pytest', 'pytest-cov', 'pytest-timeout', 'pyspark'
]

docs_require = ['Sphinx', 'sphinx-autobuild', 'sphinx-rtd-theme']

setup(
    name='mozanalysis',
    use_scm_version=True,
    author='Rob Hudson',
    author_email='robhudson@mozilla.com',
    description='A library for Mozilla experiments analysis',
    url='https://github.com/mozilla/mozanalysis',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    install_requires=[
        'attrs',
        'numpy',
        'pandas',
        'scipy',
        'google-cloud-bigquery',
    ],
    setup_requires=['pytest-runner', 'setuptools_scm'],
    extras_require={
        'testing': tests_require,
        'docs': docs_require,
    },
    tests_require=tests_require,
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import attr


@attr.s(frozen=True, slots=True)
class SegmentDataSource:
    """Represents a table or view, from which segments may be defined.

    ``window_start`` and ``window_end`` define the window of data used
    to determine whether each client fits a segment. Ideally this
    window ends at/before the moment of enrollment, so that user's
    branches can't bias the segment assignment. ``window_start`` and
    ``window_end`` are non-positive integers, representing the number
    of days before enrollment::

        window_start <= window_end <= 0


    Args:
        name (str): Name for the Data Source. Should be unique to avoid
            confusion.
        from_expr (str): FROM expression - often just a fully-qualified
            table name. Sometimes a subquery.
        window_start (int, optional): See above.
        window_end (int, optional): See above.
        client_id_column (str, optional): Name of the column that
            contains the ``client_id`` (join key). Defaults to
            'client_id'.
        submission_date_column (str, optional): Name of the column
            that contains the submission date (as a date, not
            timestamp). Defaults to 'submission_date'.
    """
    name = attr.ib(validator=attr.validators.instance_of(str))
    from_expr = attr.ib(validator=attr.validators.instance_of(str))
    window_start = attr.ib(default=0, type=int)
    window_end = attr.ib(default=0, type=int)
    client_id_column = attr.ib(default='client_id', type=str)
    submission_date_column = attr.ib(default='submission_date', type=str)

    def build_query(self, segment_list, time_limits, experiment_slug):
        """Return a nearly self contained SQL query.

        The query takes a list of ``client_id``s from
        ``raw_enrollments``, and adds one non-NULL boolean column per
        segment: True if the client is in the segment, False otherwise.
        """
        return """SELECT
            e.client_id,
            {segments}
        FROM raw_enrollments e
            LEFT JOIN {from_expr} ds
                ON ds.{client_id} = e.client_id
                AND ds.{submission_date} BETWEEN
                    DATE_ADD('{first_enrollment}', interval {window_start} day)
                    AND DATE_ADD('{last_enrollment}', interval {window_end} day)
                AND ds.{submission_date} BETWEEN
                    DATE_ADD(e.enrollment_date, interval {window_start} day)
                    AND DATE_ADD(e.enrollment_date, interval {window_end} day)
        GROUP BY e.client_id""".format(
            client_id=self.client_id_column,
            submission_date=self.submission_date_column,
            from_expr=self.from_expr,
            first_enrollment=time_limits.first_enrollment_date,
            last_enrollment=time_limits.last_enrollment_date,
            window_start=self.window_start,
            window_end=self.window_end,
            segments=',\n            '.join(
                f"{m.select_expr} AS {m.name}" for m in segment_list
            ),
        )

    @window_end.validator
    def window_end_is_not_positive(self, attribute, value):
        if value > 0:
            raise ValueError("window_end must be <= 0")

    @window_start.validator
    def window_start_lte_window_end(self, attribute, value):
        if value > self.window_end:
            raise ValueError("window_start must be <= window_end")


@attr.s(frozen=True, slots=True)
class Segment:
    """Represents an experiment Segment.

    Args:
        name (str): The segment's name; will be a column name.
        data_source (SegmentDataSource): Data source that provides
            the columns referenced in ``select_expr``.
        select_expr (str): A SQL select expression that includes
            an aggregation function (we ``GROUP BY client_id``).
            Returns a non-NULL ``BOOL``: ``True`` if the user is in the
            segment, ``False`` otherwise.
    """
    name = attr.ib(type=str)
    data_source = attr.ib(validator=attr.validators.instance_of(SegmentDataSource))
    select_expr = attr.ib(type=str)
